"""An instance-wide configuration"""
import typing as t
from threading import Lock
from collections.abc import Mapping
import logging

import yaml
import tomllib

__all__ = ("ConfigException", "Config", "Setting")

logger = logging.getLogger(__name__)


class ConfigException(Exception):
    """An exception raised in validating or modifying the config"""


# dummy placeholder object
missing = object()


class Setting:
    """A descriptor for a Config setting."""

    __slots__ = ("value", "desc", "allowed_types")

    #: The value for the setting
    value: t.Any

    #: (optional) The description for this setting
    desc: str

    #: (optional) A tuple of allowed types for the values
    allowed_types: t.Tuple[t.Any]

    #: The delimiter used for splitting keys
    delim: str = "."

    def __init__(
        self,
        value: t.Any,
        desc: str = "",
        allowed_types: t.Optional[t.Tuple[t.Any, ...]] = None,
    ):
        self.value = value
        self.desc = desc
        self.allowed_types = allowed_types

    def __set_name__(self, owner, name):
        cls_name = owner.__name__
        location = f"{cls_name}.{name}"
        self._config_insert(location)

    def __repr__(self):
        return f"Param({self.value})"

    def __get__(self, instance, owner):
        """Get the setting value"""
        return self.value

    def __set__(self, instance, value):
        raise ConfigException(
            f"Can't set Setting attribute with "
            f"value '{value}'--use the Config.update or load "
            f"methods."
        )

    def _config_insert(self, location: str):
        """Insert this setting in the config at the given location.

        Settings
        ----------
        location
            The location of the setting. e.g. 'Obj.a'

        Raises
        ------
        ConfigException
            Raised if trying to insert this setting by something else already
            exists at that location in the config.
        """
        keys = location.split(self.delim)

        # Got through each key to get sub configs deeper in the nested tree
        sub_config = Config()
        for key in keys[:-1]:
            sub_config = getattr(sub_config, key)

        # The last key is where this setting should be inserted.
        last_key = keys[-1]
        if last_key not in sub_config.__dict__:
            # If it's not inserted already, place it in there.
            sub_config.__dict__[last_key] = self
        elif id(sub_config.__dict__[last_key]) != id(self):
            # Oops, another object's in that place! There's a setting collision
            other_value = sub_config.__dict__[last_key]
            raise ConfigException(
                f"Cannot insert '{self}' at '{location}'. The '{other_value}' "
                f"already exists there."
            )


class ConfigMeta(type):
    """A metaclass to set up the creation of Config object singleton"""

    def __call__(cls, root: bool = True, **kwargs):
        # noinspection PyArgumentList
        obj = cls.__new__(cls, root=root)
        obj.__init__(**kwargs)
        return obj


class Config(metaclass=ConfigMeta):
    """A thread-safe Config manager to get and set configuration settings.

    Notes
    -----
    The base class is deliberately modified dynamically with descriptors. This
    is because the Config was designed to be configured on the fly.
    """

    #: The root singleton instance
    _instance: t.Optional["Config"] = None

    #: The thread lock
    _thread_lock: Lock = Lock()

    def __new__(cls, root: bool = True):
        # Create the singleton instance if it hasn't been created
        if cls._instance is None:
            # Lock the thread
            with cls._thread_lock:
                # Prevent another thread from creating the instance in the
                # interim
                if cls._instance is None:
                    cls._instance = super().__new__(cls)

        if not root:
            # If this is not the root Config--i.e. a sub config--return a
            # new Config object instead of the root singleton
            obj = super().__new__(cls)
            return obj
        else:
            # Otherwise return the singleton
            return cls._instance

    def __init__(self, **kwargs):
        # Set the specified settings
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattribute__(self, key):
        """Get attributes Config with support for attribute nesting"""
        try:
            # The returned value should be a setting or a Config object
            value = super().__getattribute__(key)
            return value.value if isinstance(value, Setting) else value
            # return super().__getattribute__(key)
        except AttributeError:
            # Create a dict by default for a missing attribute
            # This allows nested attribute access
            self.__dict__[key] = Config(root=False)
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        if key in ("_instance",):
            # Special keys that can have their values replaced
            super().__setattr__(key, value)
        elif key not in self.__dict__ and isinstance(value, Setting):
            # New entries are allowed as long as they are settings
            self.__dict__[key] = value
        elif not isinstance(value, Setting):
            raise ConfigException(f"Only Settings can be inserted in the Config")
        elif key in self.__dict__:
            # If it already exists, don't allow a rewrite
            raise ConfigException(
                f"Entry '{key}' already in the Config--use a "
                f"Config.update or load method to change its value."
            )
        else:
            raise ConfigException(f"Unable to chance Config")

    def __contains__(self, item):
        return item in self.__dict__

    def update(self, updates: Mapping) -> None:
        """Recursively update config with values from a dict.

        Settings
        ----------
        updates
            The dict with values to recursively update

        Raises
        ------
        ConfigException
            Raised when the updates include a change that would replace a
            sub-config (subsection) that includes settings with a single
            setting.
        KeyError
            If a key as specified that doesn't exist in this config
        ValueError
            If the update dict contains a  value that has a different
            type then the corresponding setting's value
        """
        for k, v in updates.items():
            try:
                current_value = self.__dict__[k]
            except KeyError:
                raise KeyError(
                    f"Tried assigning setting with name '{k}' which does "
                    f"not exist in the Config"
                )

            if isinstance(v, Mapping):
                # Use the corresponding update function. ex: dict.update
                current_value.update(v)
            elif isinstance(v, Config):
                # If it's a sub-config object, use its corresponding update
                # (this method)
                current_value.update(v.__dict__)
            elif isinstance(current_value, Config) and len(current_value.__dict__) > 0:
                # In this case, the update value 'v' is a simple value but the
                # current_value is a sub-config with items in it. Overwriting
                # This sub-config is not allowed
                raise ConfigException(
                    f"Cannot replace config section '{k}' with a setting '{v}'"
                )
            elif isinstance(current_value, Setting):
                # Get the setting value's type and allowed types
                value_type = type(current_value.value)
                allowed_types = current_value.allowed_types
                types = [] if allowed_types is None else list(allowed_types)
                types += [value_type]

                # Replace the setting's value, trying to coerce the type
                found_type = False
                for allowed_type in types:
                    try:
                        current_value.value = allowed_type(v)
                        found_type = True
                    except ValueError:
                        continue

                if not found_type:
                    raise ValueError(
                        f"Could not convert '{v}' into any of the "
                        f"following types: {types}"
                    )

            else:
                raise ConfigException("Setting not in Config")

    def dump(self) -> dict:
        """Convert the config into a tree dict structure"""
        d = dict()
        for k, v in self.__dict__.items():
            if isinstance(v, Setting):
                d[k] = v.value
            elif isinstance(v, Config):
                d[k] = v.dump()  # This method
            else:
                raise ConfigException(f"Unknown config type for entry '{v}'")
        return d

    # yaml methods

    @staticmethod
    def _to_yaml(value) -> str:
        """Convert a value to a string formatted for yaml"""
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return str(value)

    def loads_yaml(self, string: str) -> None:
        """Load settings from a yaml string into the config"""
        d = yaml.load(string, Loader=yaml.Loader)
        self.update(d)

    def load_yaml(self, filepath: str) -> None:
        """Load settings from a yaml file into the config"""
        with open(filepath, "r") as f:
            d = yaml.load(f, Loader=yaml.Loader)
        self.update(d)

    def dumps_yaml(self, level: int = 0, indent: int = 2) -> str:
        """Dump settings from the config into a yaml string

        Parameters
        ----------
        level
            The current level of the config for nested configs
        indent
            The number of spaces to indent each level

        Returns
        -------
        yaml
            A yaml-formatted string
        """
        string = ""
        for k, v in self.__dict__.items():
            spacer = " " * indent * level
            if isinstance(v, Setting):
                # Format the setting value and, optionally, its description
                comment = f"  # {v.desc}" if v.desc else ""
                value = self._to_yaml(v.value)
                string += f"{spacer}{k}: {value}{comment}\n"

            elif isinstance(v, Config):
                # Print the config as a new section then use the config's corresponding
                # dump method to get its values
                string += f"{spacer}{k}:\n"
                string += v.dumps_yaml(level=level + 1, indent=indent)  # this method
            else:
                # Only Config and Setting objects should be in here
                raise ConfigException(f"Unknown config type for entry '{v}'")
        return string

    # toml methods

    def loads_toml(self, string: str) -> None:
        """Load settings from a toml string into the config"""
        d = tomllib.loads(string)
        self.update(d)

    def load_toml(self, filepath: str) -> None:
        """Load settings from a toml file into the config"""
        with open(filepath, "rb") as f:
            d = tomllib.load(f)
        self.update(d)



from thatway import Setting, ConfigException

import pytest


def test_cls_attribute(config):
    """Test the setting of a class attribute"""

    class Obj:
        a = Setting(value=3)

    obj = Obj()

    assert Obj.a == 3
    assert obj.a == 3

    assert config.Obj.a == 3

    # Check the setting object
    param_obj = config.__dict__["Obj"].__dict__["a"]
    assert isinstance(param_obj, Setting)


def test_cls_attribute_mutation(config):
    """Test the mutating of a class attribute"""

    class Obj:
        a = Setting(value=3)

    obj = Obj()

    # Can't edit directly
    with pytest.raises(ConfigException):
        obj.a = 5

    # NB: This statement check's the attribute's value, and it registers the attribute
    # in the config--needed for the next test
    assert obj.a == 3

    # Can't edit the config value directly
    with pytest.raises(ConfigException):
        config.Obj.a = 5

    assert config.Obj.a == 3


def test_direct_access(config):
    """Test direct access to config"""
    # Try at the root level
    config.a = Setting("direct access")
    assert config.a == "direct access"
    assert "a" in config.__dict__

    # Try at a nested level
    config.nested.b = Setting("sub level")
    assert config.nested.b == "sub level"
    assert "b" in config.__dict__["nested"].__dict__

    # Directly setting values without a Setting is not allowed
    with pytest.raises(ConfigException):
        config.c = 3

    # Check the setting object
    param_obj = config.__dict__["a"]
    assert isinstance(param_obj, Setting)


def test_direct_access_mutation(config):
    """Test the mutating of a directly accessed config value"""
    config.a = Setting("direct access")
    config.nested.b = Setting("sub level")

    # Modifying raises an exception
    with pytest.raises(ConfigException):
        config.a = Setting("new value")

    assert config.a == "direct access"

    with pytest.raises(ConfigException):
        config.nested.b = Setting("new value")

    assert config.nested.b == "sub level"


def test_overwrite(config):
    """Test that settings cannot be overwritten in the config"""

    # 1. direct access
    config.a = Setting(1)

    with pytest.raises(ConfigException):
        config.a = Setting(2)

    assert config.a == 1

    # 2. instance attribute
    class Obj:
        b = Setting(3)

    obj = Obj()

    with pytest.raises(ConfigException):
        obj.b = 4

    assert obj.b == 3

    # 3. class attribute. These can be overwritten because the descriptor can
    # be replaced.
    # see: https://docs.python.org/3/reference/datamodel.html#object.__set__
    Obj.b = 4
    assert Obj.b == 4
    assert isinstance(Obj.__dict__["b"], int)  # not a setting anymore


def test_config_update(config):
    """Test the config.update method"""
    config.a = Setting(1)
    config.nested.b = Setting(2)

    # Updating allows overwrites
    config.update({"a": 3, "nested": {"b": 4}})

    assert config.a == 3
    assert config.nested.b == 4


def test_config_update_type_matching(config):
    """Test the config.update method with mismatched types"""
    config.a = Setting(1)
    config.nested.b = Setting(2, allowed_types=(int, str))

    # Can't change the value of 'a'
    with pytest.raises(ValueError):
        config.update({"a": "new value"})

    # Can change the value of 'nested.b' to a int or string
    config.update({"nested": {"b": "my new string"}})
    assert config.nested.b == "my new string"


def test_config_update_missing(config):
    """Test the config.update method for a missing setting, which shouldn't be
    inserted"""
    with pytest.raises(KeyError):
        config.update({"new_value": "2"})


def test_config_dump(config):
    """Test the config.dump method for converting into a dict tree"""

    # Set up the config
    class Obj:
        a = Setting(1)

    config.b = Setting("name")
    config.nested.c = Setting(True)

    # Dump and test the contents
    d = config.dump()

    assert isinstance(d, dict)
    assert d.keys() == {"Obj", "b", "nested"}

    assert isinstance(d["Obj"], dict)
    assert d["Obj"].keys() == {"a"}
    assert d["Obj"]["a"] == 1

    assert d["b"] == "name"

    assert isinstance(d["nested"], dict)
    assert d["nested"].keys() == {"c"}
    assert isinstance(d["nested"]["c"], bool)
    assert d["nested"]["c"]


@pytest.mark.parametrize("mode", ("string", "file"))
def test_config_loads_yaml(config, mode, tmp_path):
    """Test the config.loads_yaml and config.load_yaml methods for loading yaml
    strings"""
    # Set up a config
    config.a = Setting(1)

    # Load and replace the value
    config.loads_yaml("a: 2")

    # Test loading of new value
    yaml_string = "a: 2"
    if mode == "string":
        config.loads_yaml(yaml_string)
    elif mode == "file":
        tmp_file = tmp_path / "settings.yaml"
        tmp_file.write_text(yaml_string)
        config.load_yaml(tmp_file)

    assert config.a == 2


def test_config_dumps_yaml(config):
    """Test the config.dumps_yaml method for generating yaml strings"""

    # Setup a config
    class Obj:
        a = Setting(1)

    config.b = Setting("name", desc="The 'b' setting")
    config.nested.c = Setting(True)

    # Retrieve and compare the yaml string
    yaml = config.dumps_yaml()
    assert yaml == "Obj:\n  a: 1\nb: name  # The 'b' setting\nnested:\n  c: true\n"

    # The string can be loaded back without exception
    config.loads_yaml(yaml)


@pytest.mark.parametrize("mode", ("string", "file"))
def test_config_loads_toml(config, mode, tmp_path):
    """Test the config.loads_toml and config.load_toml methods for loading
    toml strings"""

    # Setup a config
    class Obj:
        a = Setting(1)

    # Test loading of new value
    toml_string = "[Obj]\na = 2"
    if mode == "string":
        config.loads_toml(toml_string)
    elif mode == "file":
        tmp_file = tmp_path / "settings.toml"
        tmp_file.write_text(toml_string)
        config.load_toml(tmp_file)

    assert Obj.a == 2
